# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
from dataclasses import dataclass
from enum import Enum, Flag, auto
from json.decoder import JSONDecodeError
from typing import Any, Mapping

import torch

from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator, INCAcceleratorType
from ..utils.logger import logger
from ..prepare_quant.prepare_model import get_world_size, get_local_rank
from .._core.scale_methods.scale_method_parser import parse_scale_method, validate_and_populate_scale_method, convert_scale_method_strings_to_enum
from .._core.scale_methods.scale_method_config import get_scale_method_from_config, check_scale_method_fields, ScaleMethodString, CfgStr, ScaleGranularity, ScaleValueType, ScaleRoundMethod


class QuantMode(Enum):
    NONE = 0
    QUANTIZE = 1
    MEASURE = 2
    SHAPE = 3
    LOAD = 4

class MeasureExclude(Flag):
    NONE = auto()
    INPUT = auto()
    OUTPUT = auto()
    PARAMS = auto()
    ALL = auto()

class SupportedFp8(Enum):
    E4M3 = torch.float8_e4m3fn
    E5M2 = torch.float8_e5m2

class HpDtype(Enum):
    BF16 = torch.bfloat16
    FP16 = torch.float16
    FP32 = torch.float32

class TrueFalse(Enum):
    TRUE = True
    FALSE = False


class ScaleFormat(Enum):
    CONST = 1  # scales is const and persistent tensor
    SCALAR = 2  # scales is non-const, non-persistent tensor with data ptr, used for low BS performance optimization


class DeviceForScalesType(Enum):
    GAUDI2 = INCAcceleratorType.GAUDI2
    GAUDI3 = INCAcceleratorType.GAUDI3


_config_to_enum = {
    "mode": QuantMode,
    "measure_exclude": MeasureExclude,
    "fp8_config": SupportedFp8,
    "hp_dtype": HpDtype,
    "scale_method": ScaleMethodString,
    "recalc_scales": TrueFalse,
    "ignore_modules_wo_measures": TrueFalse,
    "use_qdq": TrueFalse,
    "fake_quant": TrueFalse,
    "scale_format": ScaleFormat,
    "device_for_scales": DeviceForScalesType,
    "measure_on_hpu": TrueFalse,
    "dynamic_quantization": TrueFalse,
}


_configs_that_use_enum_value = [
    "fp8_config",
    "hp_dtype",
    "ignore_modules_wo_measures",
    "recalc_scales",
    "fake_quant",
    "use_qdq",
    "device_for_scales",
    "measure_on_hpu",
    "dynamic_quantization",
]

# TODO [SW-217813]: support dynamic quantization in all ops and remove
from neural_compressor.torch.algorithms.fp8_quant.model_configs import get_patched_module_table, ModuleInfo

def is_supported_dynamic_op(op_str):
    """
    Dynamically checks if the given op supports dynamic quantization
    by looking up its ModuleInfo and checking for a 'supports_dynamic_quantization' attribute.
    """
    patched_table = get_patched_module_table()
    info = patched_table.get(op_str)
    ret = getattr(info, "supports_dynamic_quantization", False) if info is not None else False
    logger.trace("Checking if %s is supported for dynamic quantization: %s", op_str, ret)
    return ret


def get_hqt_config(mod) -> Fp8cfg:
    return mod.__hqt_config__


def set_hqt_config(mod, config):
    mod.__hqt_config__ = config


def _get_enum_from_string(EnumClass, string, key):
    string = str(string)  # bool must be converted to string
    if not hasattr(EnumClass, string.upper()):
        raise ValueError(
            f"Invalid '{key}' value in custom config ('{string}'). Enter one of {[m.name for m in EnumClass]}"
        )
    return EnumClass[string.upper()]


def _validate_dump_path(dump_stats_path):
    dirname = os.path.dirname(dump_stats_path)
    basename = os.path.basename(dump_stats_path)
    if not os.access(dirname, os.W_OK):  # checks if the directory is not writable
        raise ValueError(f"Measurements dump directory '{dirname}' is non-writable")
    files_to_backup = [fname for fname in os.listdir(dirname) if fname.startswith(basename)]
    if files_to_backup:
        from datetime import datetime
        backup_dirname = f"backup_{basename}_{datetime.now().strftime('%d-%m_%H:%M:%S')}"
        try:
            os.mkdir(f"{dirname}/{backup_dirname}")
        except FileExistsError:
            pass
        for fname in files_to_backup:
            try:
                os.rename(f"{dirname}/{fname}", f"{dirname}/{backup_dirname}/{fname}")
            except (OSError, FileNotFoundError):
                pass


@dataclass
class Fp8cfg:
    cfg: Mapping[str, Any]

    def parse(custom_config: Mapping[str, str]) -> Fp8cfg:
        world_size = get_world_size()
        local_rank = get_local_rank()
        measured_global_config = {
            "dump_stats_path": "stats",
            "fp8_config": torch.float8_e4m3fn,  # The parameters of the chosen Quantization methed
            "hp_dtype": torch.bfloat16,  # The parameters of the chosen Quantization methed
            "blocklist": {
                "names": [],
                "types": (),
            },  # types and names to not be quantized
            "allowlist": {
                "names": [],
                "types": (),
            },  # types and names to be quantized. Allowlist by names is not yet implemented
            "mode": QuantMode.QUANTIZE,  # Quantize or Measure
            "fake_quant": False, # Fake or Real Quant, fake_quant only works for linear(PatchedLinear) and matmul(PatchedMatmul), usually used for training.
            "use_qdq": False, # QDQ or Real Quant, QDQ works for operators in helper_modules.py, usually used for inference.
            "scale_method": ScaleMethodString.MAXABS_HW,  # Method to quantize with
            "scale_params": {},  # scaling parameters that are different then the default ones
            "observer": "maxabs",  # Supported ['shape', 'maxabs', 'maxabs_per_channel', 'save']
            "mod_dict": {},
            "ignore_modules_wo_measures": False,  # Determines whether to fail quantization on modules without existing measures or not to quantize them
            "local_rank": local_rank if local_rank >= 0 else None,
            "global_rank": None,
            "world_size": world_size if world_size >= 0 else None,
            "seperate_measure_files": True,  # Determines whether to expect one or several measure files when using more than one gaudi
            "device_type": auto_detect_accelerator().get_inc_accelerator_type(),  # Determines device type: Gaudi2, Gaudi3...
            "device_for_scales": None,  # Overrides device type for scale: Gaudi2, Gaudi3... Enables using only G2 scales on G3
            "measure_exclude": MeasureExclude.OUTPUT,
            "recalc_scales": False,
            "scale_format": ScaleFormat.SCALAR,
            "measure_on_hpu": True,  # Determines whether to measure model on hpu device.
            "row_parallel_linear_allreduce_quantization" : False, # Turn on/off fp8 allreduce optimization detailed in SW-207602
            "dynamic_quantization" : False, # Turn on/off fp8 dynamic quantization
            "calibration_sample_interval" : 0 # number of samples to process before dumping measurements, 0 means no automatic dumping
        }
        # go over all user-defined keys from json, handle various cases
        for keys in custom_config:
            if keys in _config_to_enum.keys():
                if keys == "scale_method" and isinstance(custom_config[keys], dict):
                    # TODO: SW-230643 Add hash_id to file name in new scale method config
                    measured_global_config["recalc_scales"] = True
                    measured_global_config["scale_method"] = convert_scale_method_strings_to_enum(custom_config[keys])
                    continue
                custom_config[keys] = _get_enum_from_string(_config_to_enum[keys], custom_config[keys], keys)
                if keys in _configs_that_use_enum_value:
                    custom_config[keys] = custom_config[keys].value

            # TODO [SW-175936] - remove checking for old key names whitelist and blacklist.
            if isinstance(custom_config[keys], dict):
                for keys_2 in custom_config[keys]:
                    if keys == "whitelist":
                        measured_global_config["allowlist"][keys_2] = custom_config[keys][keys_2]
                    elif keys == "blacklist":
                        measured_global_config["blocklist"][keys_2] = custom_config[keys][keys_2]
                    else:
                        measured_global_config[keys][keys_2] = custom_config[keys][keys_2]
            else:
                if keys == "whitelist":
                    measured_global_config["allowlist"] = custom_config[keys]
                elif keys == "blacklist":
                    measured_global_config["blocklist"] = custom_config[keys]
                else:
                    measured_global_config[keys] = custom_config[keys]
        measured_global_config["scale_method"] = parse_scale_method(measured_global_config["scale_method"])
        scale_method_config = measured_global_config["scale_method"]
        validate_and_populate_scale_method(scale_method_config)


        if auto_detect_accelerator().current_device_name() == "cpu":
            if not measured_global_config["use_qdq"]:
                raise ValueError("For FP8 quantization, only QDQ mode is supported on CPU device.")
            if measured_global_config["scale_format"] == ScaleFormat.CONST and \
                check_scale_method_fields(scale_method_config, granularity_weight=ScaleGranularity.PTS, reducer=any):
                    measured_global_config["scale_format"] = ScaleFormat.SCALAR
                    logger.warning(f"FP8 per-tensor quantization on CPU device requires 'scale_format = SCALAR'")

        # If seperate_measure_files is True (default value), then it is assumed that there are multiple distinct measure and scale files
        # and they are stored in / loaded from paths with the correct index as a suffix. Else, only one is searched for.
        measured_global_config["local_rank"] = (
            local_rank if local_rank >= 0 and custom_config.get("seperate_measure_files", True) else None
        )
        # set device_for_scales config for gaudi device only
        if measured_global_config["device_type"].value > INCAcceleratorType.GAUDI_MIN.value:
            logger.debug("setting device for scales config")
            Fp8cfg.set_gaudi_device_for_scales(custom_config, measured_global_config, scale_method_config)

        if measured_global_config["scale_format"] == ScaleFormat.SCALAR:
            if check_scale_method_fields(scale_method_config, granularity_weight=ScaleGranularity.PCS, reducer=any) or \
               check_scale_method_fields(scale_method_config, granularity_activation=ScaleGranularity.PCS, reducer=any):
                measured_global_config["scale_format"] = ScaleFormat.CONST
                logger.warning(f"Cannot use 'scale_format = SCALAR' when using PCQ (Per Channel Quantization), Reduced to 'CONST'.")
            if measured_global_config["fake_quant"]:
                measured_global_config["scale_format"] = ScaleFormat.CONST
                logger.warning(f"Cannot use 'scale_format = SCALAR' when using fake_quant. Reduced to 'CONST'.")
        quant_mode = measured_global_config["mode"]
        dynamic_quantization = measured_global_config["dynamic_quantization"]
        # TODO [SW-217814]: get dynamic methods in a better way, or support file handling in dynamic mode
        if dynamic_quantization:
            if auto_detect_accelerator().current_device_name() == "cpu":
                raise ValueError("Currently CPU device doesn't support dynamic quantization")
            logger.info(f"NOTE: Using dynamic scale method, only supported ops will be quantized.")
            if measured_global_config["scale_format"] == ScaleFormat.SCALAR:
                measured_global_config["scale_format"] = ScaleFormat.CONST
                logger.warning(f"Cannot use 'scale_format = SCALAR' when using dynamic quantization. Reduced to 'CONST'.")
            # TODO: "Linear only" in types still causes issues as llama7b quantizes also self_attn,
            # which should be blocked for some reason. We might then want to set measured_global_config["allowlist"]["types"] = supported_dynamic_ops
            # TODO [SW-222725]: support HW aligned rounding in dynamic quantization
            if check_scale_method_fields(scale_method_config, rounding_method_activation= ScaleRoundMethod.HW_ALIGNED, reducer=any) or \
                check_scale_method_fields(scale_method_config, scale_value_type_activation= ScaleValueType.FIXED_VALUE, scale_value_type_weight= ScaleValueType.FIXED_VALUE, reducer=any):
                raise ValueError(
                    f"Unsupported config: scale_method {scale_method_config} is not supported in dynamic quantization"
                )
            if quant_mode not in (QuantMode.QUANTIZE, QuantMode.LOAD):
                raise ValueError(f"Quantization mode is {quant_mode}, but dynamic quantization is only supported in QUANTIZE or LOAD mode")
            measured_global_config["use_stats_files"] = False
            #TODO [SW-224403]: enable dynamic quantization in row parallel allreduce
            if measured_global_config["row_parallel_linear_allreduce_quantization"]:
                raise ValueError(f"Dynamic quantization is not supported when using row_parallel_linear_allreduce_quantization")
        else:
            if check_scale_method_fields(scale_method_config, scale_method= ScaleMethodString.ACT_MAXABS_PCS_POW2_WEIGHT_MAXABS_PTS_POW2_HW, reducer=any):
                raise ValueError(
                    f"Unsupported config: scale_method ACT_MAXABS_PCS_POW2_WEIGHT_MAXABS_PTS_POW2_HW is supported only in dynamic quantization"
                )

        if (dynamic_quantization or
            check_scale_method_fields(scale_method_config, scale_value_type_activation= ScaleValueType.FIXED_VALUE, scale_value_type_weight= ScaleValueType.FIXED_VALUE, reducer=all)) and \
            quant_mode in (QuantMode.QUANTIZE, QuantMode.LOAD):
                logger.debug(f"Quantization mode is QUANTIZE or LOAD and all scale_method are quant only so stats files won't be used")
                measured_global_config["use_stats_files"] = False
        else:
            measured_global_config["use_stats_files"] = True
            base_name = os.path.basename(measured_global_config["dump_stats_path"])
            folder_name = os.path.dirname(measured_global_config["dump_stats_path"])
            measured_global_config["dump_stats_base_path"] = folder_name
            os.makedirs(folder_name, exist_ok=True)
            worker_st = (
                ""
                if measured_global_config["local_rank"] is None
                else "_" + str(measured_global_config["local_rank"]) + "_" + str(measured_global_config["world_size"])
            )
            measured_global_config["shape_file"] = measured_global_config["dump_stats_path"] + "_hooks_shape" + worker_st
            measured_global_config["scale_file"] = (
                measured_global_config["dump_stats_path"]
                + "_hooks_"
                + measured_global_config["observer"]
                + "_"
                + get_scale_method_from_config(measured_global_config["scale_method"][CfgStr.DEFAULT]).name #TODO SW-230643 Add hash_id to file name in new scale method config
                + worker_st
            )
            if (quant_mode == QuantMode.MEASURE) or (
                quant_mode == QuantMode.QUANTIZE
            ):
                measured_global_config["measure_file"] = (
                    measured_global_config["dump_stats_path"] + "_hooks_" + measured_global_config["observer"] + worker_st
                )
            # measured_global_config['dump_stats_path'] += '_hooks_.json'

            logger.debug("HQT Paths:")
            logger.debug("base_name='%s'", base_name)
            logger.debug("folder_name='%s'", folder_name)
            logger.debug(
                "measured_global_config['shape_file']='%s'",
                measured_global_config["shape_file"],
            )
            logger.debug(
                "measured_global_config['scale_file']='%s'",
                measured_global_config["scale_file"],
            )
            if "measure_file" in measured_global_config.keys():
                logger.debug(
                    "measured_global_config['measure_file']='%s'",
                    measured_global_config["measure_file"],
                )
            logger.debug(
                "measured_global_config['dump_stats_path']='%s'",
                measured_global_config["dump_stats_path"],
            )

        if measured_global_config["mode"] == QuantMode.MEASURE:
            _validate_dump_path(measured_global_config["dump_stats_path"])

        return Fp8cfg(cfg=measured_global_config)

    @staticmethod
    def set_gaudi_device_for_scales(custom_config, measured_global_config, scale_method):
        current_device_type = measured_global_config["device_type"]
        if current_device_type.value < INCAcceleratorType.GAUDI_MIN.value:
            raise ValueError("device for scales config is supported for only gaudi device line.")
        if custom_config.get("device_for_scales", None) is None:
            # Device for scales is the current device by default
            measured_global_config["device_for_scales"] = current_device_type

        elif measured_global_config["device_for_scales"] != measured_global_config["device_type"]:
            # Currently, only maxabs_hw is supported for a different device scales configuration
            if not check_scale_method_fields(scale_method_dict=scale_method, scale_method= ScaleMethodString.MAXABS_HW, reducer=all):
                raise ValueError(
                    f"Unsupported config: scale_method: {scale_method} for scale device overriding: {measured_global_config['device_for_scales']}"
                )
            if not (
                measured_global_config["device_for_scales"] == INCAcceleratorType.GAUDI2
                and measured_global_config["device_type"] == INCAcceleratorType.GAUDI3
            ):
                raise ValueError(f"Unsupported config: device_for_scales={measured_global_config['device_for_scales']} "
                                f"for device_type={measured_global_config['device_type']}")

def _read_config_from_file(config_path: str) -> Mapping[str, str]:
    logger.debug("QUANT PACKAGE: using %s config", config_path)

    module_directory = os.path.dirname(os.path.abspath(__file__))

    # if file in absolute path doesn't exist, try looking in cfg directory
    if not os.path.isfile(config_path):
        config_path = os.path.join(module_directory, "..", f"custom_config/{config_path}")
    try:
        logger.info("QUANT PACKAGE: Loading %s", config_path)
        with open(config_path) as config_json:
            config = json.load(config_json)
    except FileNotFoundError as e:
        raise Exception(f"Got exception: {e}. QUANT PACKAGE: Can't open {config_path}!")
    except JSONDecodeError as e:
        config_json.close()
        raise Exception(f"Got exception: {e}. QUANT PACKAGE: Can't load {config_path}!")
    return config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch
import json
import os
import random
import re
import time

import bigcode_eval  # pylint: disable=E0611, E0401
import numpy as np
import torch
from accelerate import Accelerator
from bigcode_eval.arguments import EvalArguments  # pylint: disable=E0611, E0401
from bigcode_eval.evaluator import Evaluator  # pylint: disable=E0611, E0401
from bigcode_eval.tasks import ALL_TASKS  # pylint: disable=E0611, E0401


def pattern_match(patterns, source_list):
    """Returns a list containing all values of the source_list that
    match at least one of the patterns."""
    task_names = set()
    for pattern in patterns:
        for matching in fnmatch.filter(source_list, pattern):
            task_names.add(matching)
    return list(task_names)


def evaluate(
    model,
    tokenizer,
    tasks,
    batch_size,
    args,
):
    """Instantiate and evaluate a model on a list of tasks."""
    try:
        import datasets
        import transformers

        transformers.logging.set_verbosity_error()
        datasets.logging.set_verbosity_error()
    except:
        pass
    if tasks is None:
        task_names = ALL_TASKS
    else:
        task_names = pattern_match(tasks.split(","), ALL_TASKS)

    accelerator = Accelerator()
    if accelerator.is_main_process:
        print(f"Selected Tasks: {task_names}")

    results = {}

    if not tokenizer.eos_token:
        if tokenizer.bos_token:
            tokenizer.eos_token = tokenizer.bos_token
            print("bos_token used as eos_token")
        else:
            raise ValueError("No eos_token or bos_token found")

    try:
        tokenizer.pad_token = tokenizer.eos_token

    # Some models like CodeGeeX2 have pad_token as a read-only property
    except AttributeError:
        print("Not setting pad_token to eos_token")
        pass

    evaluator = Evaluator(accelerator, model, tokenizer, args)

    if args.load_generations_intermediate_paths and len(args.load_generations_intermediate_paths) != len(task_names):
        raise ValueError(
            "If passing --load_generations_intermediate_paths, \
            must pass equal number of files as number of tasks"
        )

    for idx, task in enumerate(task_names):
        intermediate_generations = None
        if args.load_generations_intermediate_paths:
            with open(args.load_generations_intermediate_paths[idx], "r") as f_in:
                # intermediate_generations: list[list[str | None]] of len n_tasks
                # where list[i] = generated codes or empty
                intermediate_generations = json.load(f_in)

        if args.generation_only:
            if accelerator.is_main_process:
                print("generation mode only")
            generations, references = evaluator.generate_text(task, intermediate_generations=intermediate_generations)
            if accelerator.is_main_process:
                save_generations_path = f"{os.path.splitext(args.save_generations_path)[0]}_{task}.json"
                save_references_path = f"references_{task}.json"
                evaluator.save_json_files(
                    generations,
                    references,
                    save_generations_path,
                    save_references_path,
                )
        else:
            results[task] = evaluator.evaluate(task, intermediate_generations=intermediate_generations)

    # Save all args to config
    results["config"] = vars(args)
    if not args.generation_only:
        dumped = json.dumps(results, indent=2)
        if accelerator.is_main_process:
            print(dumped)

        with open(args.metric_output_path, "w") as f:
            f.write(dumped)

    return results



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AutoRound quantization."""
import copy
import json
import time
from functools import lru_cache
from typing import Iterable, Optional, Union

import torch


@lru_cache(None)
def _is_auto_round_available():
    """Check whether the AutoRound package is importable.

    Returns:
        bool: True when the auto_round package can be imported.
    """
    try:
        import auto_round  # pylint: disable=E0401
    except ImportError:
        logger.error("AutoRound is not installed. Please install it using 'pip install auto_round'.")
        return False
    return True


_is_auto_round_available()

from auto_round import AutoRound, AutoRoundMLLM  # pylint: disable=E0401
from auto_round.compressors.mllm.eval import lmms_eval, mllm_eval
from auto_round.compressors.mllm.template import Template, get_template
from auto_round.schemes import QuantizationScheme

from neural_compressor.common.utils import Statistics
from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.algorithms.weight_only.utility import CapturedDataloader, InputCaptureModule
from neural_compressor.torch.utils import get_accelerator, logger


class AutoRoundQuantizer(Quantizer):
    """AutoRound Quantizer."""

    def __init__(
        self,
        quant_config: Optional[dict] = None,
        **kwargs,
    ):
        """Init a AutQRoundQuantizer object.

        Args:
            quant_config (dict): Configuration for weight quantization (default is None).
            quant_config={
                        'layer1':##layer_name
                        {
                            'data_type': 'int',
                            'bits': 4,
                            'group_size': 32,
                            'sym': False,
                            'act_data_type': None,
                            'act_bits': 32,
                            'act_sym': None,
                            'act_dynamic': True,
                        }
                        ...,
                    }
                keys:
                    data_type (str): The data type to be used (default is "int").
                    bits (int): Number of bits for quantization (default is 4).
                    group_size (int): Size of the quantization group (default is 128).
                    sym (bool): Whether to use symmetric quantization. (default is None).
            layer_config (dict, optional): Layer-wise quantization config. Defaults to None.
            bits (int): Number of bits for quantization (default is 4).
            group_size (int): Size of the quantization group (default is 128).
            sym (bool): Whether symmetric quantization is to be used (default is False).
            enable_full_range (bool): Whether to enable full range quantization (default is False).
            batch_size (int): Batch size for training (default is 8).
            amp (bool): Whether to use automatic mixed precision (default is True).
            device_map: The device to be used for tuning (default is None).
            quant_lm_head (bool): Indicates whether quantize the lm_head layer in transformers. (default is False).
            lr_scheduler: The learning rate scheduler to be used.
            dataset (str): The default dataset name (default is "NeelNanda/pile-10k").
            enable_quanted_input (bool): Whether to use the output of the previous quantized block as
                                            the input for the current block (default is True).
            enable_minmax_tuning (bool): Whether to enable weight min-max tuning (default is True).
            lr (float): The learning rate (default is None, will be set to 1.0/iters).
            minmax_lr (float): The learning rate for min-max tuning
                                    (default is None, it will be set to lr automatically).
            low_gpu_mem_usage (bool): Whether to use low GPU memory (default is True).
            iters (int): Number of iterations (default is 200).
            seqlen (int): Data length of the sequence for tuning (default is 2048).
            nsamples (int): Number of samples (default is 128).
            sampler (str): The sampling method (default is "rand").
            seed (int): The random seed (default is 42).
            nblocks (int): Number of blocks (default is 1).
            gradient_accumulate_steps (int): Number of gradient accumulation steps (default is 1).
            not_use_best_mse (bool): Whether to use mean squared error (default is False).
            dynamic_max_gap (int): The dynamic maximum gap (default is -1).
            data_type (str): The data type to be used (default is "int").
            scale_dtype (str): The data type of quantization scale to be used (default is "float16"), different kernels
                have different choices.
            to_quant_block_names (list): A list whose elements are list of block's layer names to be quantized.
            act_bits (int): Number of bits for activation quantization. Default is 32.
            act_group_size (int): Group size for activation quantization. Default is None.
            act_sym (bool): Whether to use symmetric activation quantization. Default is None.
            act_dynamic (bool): Whether to use dynamic activation quantization. Default is True.
            enable_norm_bias_tuning (bool): Whether to enable fast norm/layer_bias tuning.
            enable_torch_compile (bool): Whether to enable torch compile to optimize quant_block/layer, torch>=2.6 True.
            quant_nontext_module (bool): Whether to quantize nontext module.
            extra_data_dir (str): The path for extra data such as images, audio or videos.
            processor (transformers.AutoProcessor): Any multi-modal model will require an object to encode or
                decode the data that groups several modalities (among text, vision and audio).
                This is handled by objects called processors, which group together two or more processing objects such
                as tokenizers (for the text modality), image processors (for vision) and feature extractors (for audio).
            image_processor (Processor): Image processor for special model like llava.
            template (Template): The template to specify process for different mllms.
            truncation (bool): Activates truncation to cut input sequences longer than `max_length` to `max_length`.
            scheme (str| dict | QuantizationScheme ): A preset scheme that defines the quantization configurations.
            guidance_scale (float): Control how much the image generation process follows the text prompt.
                                    The more it is, the more closely it follows the prompt (default is 7.5).
            num_inference_steps (int): The reference number of denoising steps (default is 50).
            generator_seed (int): A seed that controls the initial noise for image generation (default is None).

        Returns:
            The quantized model.
        """
        super().__init__(quant_config=quant_config)
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.accelerator = get_accelerator(kwargs.pop("device", "auto"))
        self.device = self.accelerator.name()

    def _is_w4afp8(self) -> bool:
        """Return whether the configuration requests W4AFP8 quantization.

        Returns:
            bool: True when using fp8_to_int_sym data type.
        """
        return self.data_type == "fp8_to_int_sym"

    def prepare(self, model: torch.nn.Module, *args, **kwargs):
        """Prepares a given model for quantization.

        Args:
            model (torch.nn.Module): The model to be prepared.

        Returns:
            A prepared model.
        """
        prepare_model = InputCaptureModule(model)
        return prepare_model

    def convert(self, model: torch.nn.Module, *args, **kwargs):
        """Convert the prepared model to a quantized model.

        Args:
            model (torch.nn.Module): the prepared model

        Returns:
            The quantized model.
        """
        pipe = kwargs.pop("pipeline", None)
        tokenizer = getattr(model.orig_model, "tokenizer", None)
        if tokenizer is not None:
            delattr(model.orig_model, "tokenizer")
        elif pipe is None:
            tokenizer = "Placeholder"
            self.dataset = CapturedDataloader(model.args_list, model.kwargs_list)
        model = model.orig_model
        if pipe is not None:
            model = pipe
        # Remove AutoRound specific args before passing to AutoRound constructor
        keys_to_pop = ["quant_config", "device", "export_format", "output_dir", "accelerator", "reloading"]
        if hasattr(self, "target_bits") and self.target_bits is not None:
            from auto_round import AutoScheme

            self.scheme = AutoScheme(
                avg_bits=self.target_bits,
                options=self.options,
                shared_layers=self.shared_layers,
                ignore_scale_zp_bits=self.ignore_scale_zp_bits,
                method=self.auto_scheme_method,
                batch_size=self.auto_scheme_batch_size,
                device_map=self.auto_scheme_device_map,
                low_gpu_mem_usage=self.low_gpu_mem_usage,
            )
        # Remove AutoRound specific AutoScheme args before passing to AutoRound constructor
        keys_to_pop += [
            "target_bits",
            "options",
            "shared_layers",
            "ignore_scale_zp_bits",
            "auto_scheme_method",
            "auto_scheme_batch_size",
            "auto_scheme_device_map",
        ]

        rounder = AutoRound(
            model,
            tokenizer=tokenizer,
            **{k: v for k, v in self.__dict__.items() if k not in keys_to_pop},
        )

        if self._is_w4afp8():
            model, weight_config = rounder.quantize()
            model.autoround_config = weight_config
            return rounder.save_quantized(output_dir=self.output_dir, inplace=True)
        else:  # pragma: no cover
            rounder.quantize_and_save(output_dir=self.output_dir, format=self.export_format, inplace=True)
            model = rounder.model
            model.autoround_config = rounder.layer_config

        self.accelerator.empty_cache()
        dump_model_op_stats(rounder.layer_config)

        reloading = self.__dict__.get("reloading", True)
        if self.export_format in ["auto_round", "llm_compressor"] and reloading:
            # the directly returned model is QuantLinear, which is used for packing.
            try:
                logger.info(
                    f"Quantization is done, reloading model from saved directory({self.output_dir})...\n"
                    "Set reloading=False to skip."
                )
                import transformers  # pylint: disable=E0401

                model = transformers.AutoModelForCausalLM.from_pretrained(self.output_dir)
            except:
                pass

        return model


def get_dataloader(tokenizer, seqlen, dataset_name="NeelNanda/pile-10k", seed=42, bs=8, nsamples=128):
    """Generate a DataLoader for calibration using specified parameters.

    Args:
        tokenizer (Tokenizer): The tokenizer to use for tokenization.
        seqlen (int): The exact sequence length. Samples shorter than `seqlen` will be dropped,
            and samples longer than `seqlen` will be truncated.
        dataset_name (str, optional): The name of the dataset or datasets separated by commas.
            Defaults to "NeelNanda/pile-10k".
        seed (int, optional): The random seed for reproducibility. Defaults to 42.
        bs (int, optional): The batch size. Defaults to 8.
        nsamples (int, optional): The total number of samples to include. Defaults to 128.

    Returns:
        DataLoader: The DataLoader for the calibrated dataset.
    """
    from auto_round.calib_dataset import get_dataloader  # pylint: disable=E0401

    dataloader = get_dataloader(tokenizer, seqlen, dataset_name=dataset_name, seed=seed, bs=bs, nsamples=nsamples)
    return dataloader


def get_mllm_dataloader(
    model,
    tokenizer,
    template=None,
    processor=None,
    image_processor=None,
    dataset=None,
    extra_data_dir=None,
    seqlen=None,
    batch_size=8,
    split=None,
    apply_template=None,
    truncation=None,
    seed=42,
    nsamples=128,
    gradient_accumulate_steps=1,
    quant_nontext_module=False,
):
    """Generate a DataLoader for calibration using specified parameters.

    Args:
        model (Model): The model to quantize.
        tokenizer (Tokenizer): The tokenizer to use for tokenization.
        template (Template, optional): The template to specify process for different MLLMs.
        processor (transformers.AutoProcessor, optional): The processor for multi-modal inputs.
        image_processor (object, optional): The image processor for multi-modal inputs.
        dataset (str, optional): The name or path of the dataset.
        extra_data_dir (str, optional): The path for extra data such as images, audio, or videos.
        seqlen (int, optional): The exact sequence length. Samples shorter than `seqlen` will be dropped,
            and samples longer than `seqlen` will be truncated.
        batch_size (int, optional): The batch size. Defaults to 8.
        split (str, optional): The data split to use. Defaults to None.
        apply_template (bool, optional): Whether to apply chat template in tokenization.
        truncation (bool, optional): Whether to truncate sequences during tokenization.
        seed (int, optional): The random seed for reproducibility. Defaults to 42.
        nsamples (int, optional): The total number of samples to include. Defaults to 128.
        gradient_accumulate_steps (int, optional): The number of gradient accumulation steps. Defaults to 1.
        quant_nontext_module (bool, optional): Whether to quantize non-text modules. Defaults to False.

    Returns:
        DataLoader: The DataLoader for the calibrated datasets.
    """
    from auto_round.calib_dataset import CALIB_DATASETS
    from auto_round.compressors.mllm.compressor import _only_text_test
    from auto_round.compressors.mllm.dataset import get_mllm_dataloader  # pylint: disable=E0401

    template = template if template is not None else model.config.model_type
    template = get_template(
        template, model=model, tokenizer=tokenizer, processor=processor, image_processor=image_processor
    )
    dataset = template.default_dataset if dataset is None else dataset
    if quant_nontext_module or (
        dataset in CALIB_DATASETS.keys() and not _only_text_test(model, tokenizer, "cpu", template.model_type)
    ):
        if quant_nontext_module:
            logger.warning(
                "Quantitative nontext module is not supported for plain text datasets,"
                "will use liuhaotian/llava_conv_58k with default config as an alternative."
            )
        else:
            logger.warning(
                f"{model.config.model_type} not support for {dataset},"
                " will use liuhaotian/llava_conv_58k with default config as an alternative."
            )
        dataset = "liuhaotian/llava_conv_58k"
        seqlen = 512 if seqlen is None else seqlen
        truncation = False
        gradient_accumulate_steps = batch_size * gradient_accumulate_steps
        batch_size = 1
        seed = 42  # The seed is fixed to 42 in transformers
    seqlen = 2048 if seqlen is None else seqlen  # set text only calibration default args
    truncation = True if truncation is None else truncation
    dataset = dataset.replace(" ", "")

    if nsamples % batch_size != 0:
        nsamples = (nsamples // batch_size + 1) * batch_size
        logger.warning(f"'nsamples' is not divisible by 'batch_size', will adjusted to {nsamples}")

    dataloader, batch_size, seqlen, gradient_accumulate_steps = get_mllm_dataloader(
        template=template,
        processor=processor,
        model=model,
        tokenizer=tokenizer,
        image_processor=image_processor,
        dataset=dataset,
        extra_data_dir=extra_data_dir,
        seqlen=seqlen,
        bs=batch_size,
        seed=seed,
        truncation=truncation,
        nsamples=nsamples,
        gradient_accumulate_steps=gradient_accumulate_steps,
        quant_nontext_module=quant_nontext_module,
    )
    return dataloader, template, truncation, batch_size, gradient_accumulate_steps, seqlen, nsamples


def dump_model_op_stats(layer_config):
    """Dump quantizable ops stats of model to user."""
    # TODO: collect more ops besides Linear
    res = {}
    res["Linear"] = {}
    for name, info in layer_config.items():
        if "data_type" in info:
            data_type_str = info["data_type"].upper()
            if "bits" in info and str(info["bits"]) not in info["data_type"]:
                data_type_str += str(info["bits"])
            res["Linear"][data_type_str] = res.get("Linear", {}).get(data_type_str, 0) + 1

    # update stats format for dump.
    field_names = ["Op Type", "Total"]
    dtype_list = list(res["Linear"].keys())
    field_names.extend(dtype_list)
    output_data = []
    for op_type in res.keys():
        field_results = [op_type, sum(res[op_type].values())]
        field_results.extend([res[op_type][dtype] for dtype in dtype_list])
        output_data.append(field_results)

    Statistics(output_data, header="Mixed Precision Statistics", field_names=field_names).print_stat()



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor model convert."""

import json
import math
import os
import re
import types

from datasets import load_dataset

from neural_compressor.common.utils import LazyImport, logger
from neural_compressor.torch.algorithms.weight_only.modules import INCWeightOnlyLinear
from neural_compressor.torch.algorithms.weight_only.utility import repack_awq_to_optimum_format
from neural_compressor.torch.quantization import (
    AWQConfig,
    GPTQConfig,
    RTNConfig,
    TEQConfig,
    convert,
    prepare,
)
from neural_compressor.torch.utils import get_accelerator, is_ipex_available, is_package_available

if is_ipex_available():
    import intel_extension_for_pytorch as ipex

from typing import Union

torch = LazyImport("torch")


def convert_dtype_str2torch(str_dtype):
    if str_dtype == "int8":
        return torch.int8
    elif str_dtype == "fp32" or str_dtype == "auto":
        return torch.float
    elif str_dtype == "fp16":
        return torch.float16
    elif str_dtype == "bf16":
        return torch.bfloat16
    else:
        assert False, "Unsupported str dtype {} to torch dtype".format(str_dtype)


def convert_dtype_torch2str(dtype):
    if dtype == torch.int8:
        return "int8"
    elif dtype == torch.float:
        return "fp32"
    elif dtype == torch.float16:
        return "fp16"
    elif dtype == torch.bfloat16:
        return "bf16"
    elif isinstance(dtype, str) and dtype in ["int8", "fp32", "fp16", "bf16"]:
        return dtype
    else:
        assert False, "Unsupported pytorch dtype {} to str dtype".format(dtype)


def replace_linear(
    model,
    modules_to_not_convert=None,
    current_key_name=None,
    quantization_config=None,
    device="cpu",
    empty_weights=False,
):
    if modules_to_not_convert is None:
        # output_layer is chatglm last layer name
        # embed_out is dolly_v2 last layer name
        modules_to_not_convert = []
    if quantization_config.modules_to_not_convert:
        modules_to_not_convert.extend(quantization_config.modules_to_not_convert)
        modules_to_not_convert = list(set(modules_to_not_convert))
    model, is_replaced = _replace_linear(
        model,
        modules_to_not_convert,
        current_key_name,
        quantization_config,
        device=device,
        empty_weights=empty_weights,
    )

    if not is_replaced:
        logger.warning(
            "You are loading your model in 8bit or 4bit but no linear modules were found in your model."
            " Please double check your model architecture, or submit an issue on github if you think this is"
            " a bug."
        )

    return model


def _replace_linear(
    model,
    modules_to_not_convert=None,
    current_key_name=None,
    quantization_config=None,
    is_replaced=False,
    device="cpu",
    empty_weights=False,
):
    """Private method that wraps the recursion for module replacement.

    Returns the converted model and a boolean that indicates if the conversion has been successfully or not.
    """
    for name, module in model.named_children():
        if current_key_name is None:
            current_key_name = []
        current_key_name.append(name)
        is_removed = False
        if (isinstance(module, torch.nn.Linear) or isinstance(module, INCWeightOnlyLinear)) and (
            name not in modules_to_not_convert
        ):
            # Check if the current key is not in the `modules_to_not_convert`
            if not any(key in ".".join(current_key_name) for key in modules_to_not_convert) and not any(
                re.match(pattern, ".".join(current_key_name)) for pattern in modules_to_not_convert
            ):
                in_features = module.in_features
                out_features = module.out_features
                if device == "cpu" or device == torch.device("cpu") or device == "auto":
                    from intel_extension_for_pytorch.nn.modules import WeightOnlyQuantizedLinear as ipex_linear
                    from intel_extension_for_pytorch.utils.weight_only_quantization import (
                        _convert_optimum_format_to_desired,
                    )

                    qweight = module.qweight
                    scales = module.scales
                    qzeros = module.qzeros

                    qweight, scales, qzeros = _convert_optimum_format_to_desired(qweight, scales, qzeros)
                    weight_dtype = {
                        4: ipex.quantization.WoqWeightDtype.INT4,
                        8: ipex.quantization.WoqWeightDtype.INT8,
                    }
                    compute_dtype = {
                        "fp32": ipex.quantization.WoqLowpMode.NONE,  # follow the activation datatype.
                        "bf16": ipex.quantization.WoqLowpMode.BF16,
                        "fp16": ipex.quantization.WoqLowpMode.FP16,
                        "int8": ipex.quantization.WoqLowpMode.INT8,
                    }
                    ipex_qconfig_mapping = ipex.quantization.get_weight_only_quant_qconfig_mapping(
                        weight_dtype=weight_dtype[quantization_config.bits],
                        lowp_mode=compute_dtype[quantization_config.compute_dtype],
                        act_quant_mode=ipex.quantization.WoqActQuantMode.PER_IC_BLOCK,
                        group_size=quantization_config.group_size,
                    )
                    tmp_linear = torch.nn.Linear(
                        in_features,
                        out_features,
                        True if hasattr(module, "bias") and module.bias is not None else False,
                    )
                    if tmp_linear.bias is not None and module.bias is not None:
                        tmp_linear.bias = torch.nn.Parameter(module.bias.float())

                    tmp_linear.qconfig = ipex_qconfig_mapping.global_qconfig
                    model._modules[name] = ipex_linear.from_float_and_int4_weight(
                        mod=tmp_linear,
                        qweight=qweight,
                        scales=scales,
                        zero_points=qzeros,
                        bias=(module.bias.float() if hasattr(module, "bias") and module.bias is not None else None),
                        group_size=quantization_config.group_size,
                        g_idx=(module.g_idx if hasattr(module, "g_idx") else None),
                    )

                elif device == "xpu" or device == torch.device("xpu"):
                    from intel_extension_for_pytorch.nn.utils._quantize_convert import (
                        WeightOnlyQuantizedLinear as ipex_linear,  # pylint: disable=E0401
                    )

                    model._modules[name] = ipex_linear(
                        in_features,
                        out_features,
                        module.bias is not None,
                        compute_dtype=quantization_config.compute_dtype,
                        compress_statistics=False,
                        weight_dtype=quantization_config.weight_dtype,
                        scale_dtype=quantization_config.scale_dtype,
                        blocksize=quantization_config.group_size,
                        scheme=quantization_config.scheme,
                        compression_dtype=getattr(module, "compression_dtype", torch.int32),
                        compression_dim=getattr(module, "compression_dim", 1),
                        device=device,
                        use_optimum_format=getattr(module, "use_optimum_format", True),
                    )
                    # g_idx is only present when using GPTQ quantization method
                    g_idx = module.g_idx if hasattr(module, "g_idx") else None
                    model._modules[name].set_scales_zps_gidx(
                        (
                            module.scales
                            if hasattr(module, "scales")
                            else torch.ones(
                                (
                                    math.ceil(in_features / quantization_config.group_size),
                                    out_features,
                                ),
                                dtype=convert_dtype_str2torch(quantization_config.compute_dtype),
                                device=torch.device(device),
                            )
                        ),
                        module.qzeros if hasattr(module, "qzeros") else None,
                        g_idx,
                    )
                    if not hasattr(module, "qweight"):
                        n_pack = 32 // quantization_config.bits

                        weight = torch.zeros(
                            (math.ceil(in_features / n_pack), out_features),
                            dtype=torch.int32,
                            device=torch.device(device),
                        )

                    # Note: update_g_idx is only applicable for ipex versions >=2.7
                    model._modules[name].set_weights_bias(
                        module.qweight.data if hasattr(module, "qweight") else weight,
                        None if module.bias is None else module.bias.data,
                    )
                else:
                    raise Exception("{} device Unsupported weight only quantization!".format(device))

                is_replaced = True
                is_removed = True
                # Store the module class in case we need to transpose the weight later
                model._modules[name].source_cls = type(module)
                # Force requires grad to False to avoid unexpected errors
                model._modules[name].requires_grad_(False)

        if not is_removed and len(list(module.children())) > 0:  # pylint: disable=E1101
            _, is_replaced = _replace_linear(
                module,
                modules_to_not_convert,
                current_key_name,
                quantization_config,
                is_replaced=is_replaced,
                device=device,
                empty_weights=empty_weights,
            )
        # Remove the last key for recursion
        current_key_name.pop(-1)
    return model, is_replaced


def default_run_fn(model, tokenizer, dataset, max_length=512, n_samples=100, batch_size=8, algo="rtn"):
    from torch.utils.data import DataLoader

    if isinstance(dataset, (str, bytes, os.PathLike)):
        calib_dataset = load_dataset(dataset, split="train")
    calib_dataset = calib_dataset.shuffle(seed=42)
    if tokenizer is None:
        logger.error("Please provide the tokenizer in quantization_config.")
        exit(0)

    def tokenize_function(examples):
        if algo == "teq":
            if tokenizer.pad_token is None:
                tokenizer.pad_token = tokenizer.eos_token
        if "prompt" in examples:
            if algo == "teq":
                example = tokenizer(examples["prompt"], padding="max_length", max_length=max_length)
            else:
                example = tokenizer(examples["prompt"])
        elif "code" in examples:
            if algo == "teq":
                example = tokenizer(examples["code"], padding="max_length", max_length=max_length)
            else:
                example = tokenizer(examples["code"])
        elif "text" in examples:
            if algo == "teq":
                example = tokenizer(examples["text"], padding="max_length", max_length=max_length)
            else:
                example = tokenizer(examples["text"])
        else:
            logger.error(
                "Please check dataset prompt identifier," + " NeelNanda/pile-10k is default used calibration dataset."
            )
            exit(0)
        return example

    tokenized_dataset = calib_dataset.map(tokenize_function, batched=True)
    tokenized_dataset.set_format(type="torch", columns=["input_ids"])
    tokenized_dataset = tokenized_dataset.filter(lambda x: x["input_ids"].shape[-1] >= max_length)

    def collate_batch(batch):
        input_ids_padded = []
        for text in batch:
            input_ids = text["input_ids"]
            if len(input_ids) >= max_length:
                input_ids = input_ids[:max_length]
                input_ids_padded.append(input_ids)
            else:
                continue
        assert (
            input_ids_padded != []
        ), "The dataset does not have data that meets the required input length. Please reduce seq_len."
        return torch.vstack(input_ids_padded)

    calib_dataloader = DataLoader(
        tokenized_dataset,
        batch_size=batch_size,
        shuffle=False,
        collate_fn=collate_batch,
    )
    total_cnt = 0

    device = get_accelerator().current_device_name()
    for i, (input_ids) in enumerate(calib_dataloader):
        input_ids = input_ids.to(torch.device(device))
        if total_cnt + input_ids.shape[0] > n_samples:
            input_ids = input_ids[: n_samples - total_cnt, ...]
        total_cnt += input_ids.shape[0]
        if total_cnt >= n_samples:
            break

        try:
            model(
                input_ids=input_ids,
            )
        except ValueError:
            pass


def convert_to_quantized_model(model, config, device="cpu", for_inference=True):
    if device == "xpu" or device == torch.device("xpu"):
        import intel_extension_for_pytorch

        assert hasattr(torch, "xpu") and torch.xpu.is_available(), "There is no xpu device in this system!"
        if "INC_TARGET_DEVICE" not in os.environ:
            os.environ["INC_TARGET_DEVICE"] = "cpu"
            logger.info(
                "Set the environment variable INC_TARGET_DEVICE='cpu'"
                " to ensure the quantization process occurs on the CPU."
            )

    # mapping to INC config
    dtype = "int4" if config.weight_dtype == "int4_fullrange" else config.weight_dtype
    if config.quant_method.value == "rtn":
        quant_config = RTNConfig(
            dtype=dtype,
            bits=config.bits,
            use_sym=config.sym,
            group_size=config.group_size,
            use_layer_wise=config.use_layer_wise,
            model_path=config.model_path,
            quant_lm_head=config.quant_lm_head,
        )
        if config.modules_to_not_convert != []:
            for module in config.modules_to_not_convert:
                module_name = ".*" + module
                quant_config.set_local(module_name, RTNConfig(dtype="fp32"))
        logger.info(f"Do RTN algorithm with config {quant_config}")
        model = prepare(model, quant_config)
        model = convert(model)
    elif config.quant_method.value == "gptq":
        model.seqlen = config.seq_len
        quant_config = GPTQConfig(
            dtype=dtype,
            bits=config.bits,
            use_sym=config.sym,
            group_size=config.group_size,
            use_layer_wise=config.use_layer_wise,
            model_path=config.model_path,
            act_order=config.desc_act,
            percdamp=config.damp_percent,
            block_size=config.blocksize,
            static_groups=config.static_groups,
            use_mse_search=config.use_mse_search,
            true_sequential=config.true_sequential,
            quant_lm_head=config.quant_lm_head,
        )
        if config.modules_to_not_convert != []:
            for module in config.modules_to_not_convert:
                module_name = ".*" + module
                quant_config.set_local(module_name, GPTQConfig(dtype="fp32"))
        logger.info(f"Do GPTQ algorithm with config {quant_config}")
        run_fn = default_run_fn
        run_args = (
            config.tokenizer,
            config.dataset,
            config.seq_len,  # max_length
            config.n_samples,  # n_samples
            config.batch_size,  # batch_size
            config.quant_method.value,  # algo
        )
        model = prepare(model=model, quant_config=quant_config)
        run_fn(model, *run_args)
        model = convert(model)
    elif config.quant_method.value == "awq":
        quant_config = AWQConfig(
            dtype=dtype,
            bits=config.bits,
            use_sym=config.sym,
            group_size=config.group_size,
            use_layer_wise=config.use_layer_wise,
            use_auto_scale=config.auto_scale,
            use_auto_clip=config.auto_clip,
            folding=True,
            absorb_layer_dict=config.absorb_layer_dict,
            quant_lm_head=config.quant_lm_head,
        )
        if config.modules_to_not_convert != []:
            for module in config.modules_to_not_convert:
                module_name = ".*" + module
                quant_config.set_local(module_name, AWQConfig(dtype="fp32"))
        logger.info(f"Do AWQ algorithm with config {quant_config}")
        run_fn = default_run_fn
        run_args = (
            config.tokenizer,
            config.dataset,
            config.seq_len,  # max_length
            config.n_samples,  # n_samples
            config.batch_size,  # batch_size
            config.quant_method.value,  # algo
        )
        example_inputs = torch.ones([1, 512], dtype=torch.long).to(device)
        model = prepare(model=model, quant_config=quant_config, example_inputs=example_inputs)
        run_fn(model, *run_args)
        model = convert(model)
    elif config.quant_method.value == "teq":
        quant_config = TEQConfig(
            dtype=dtype,
            bits=config.bits,
            use_sym=config.sym,
            group_size=config.group_size,
            use_layer_wise=config.use_layer_wise,
            quant_lm_head=config.quant_lm_head,
            folding=config.folding,
            absorb_to_layer=config.absorb_layer_dict,
        )
        if config.modules_to_not_convert != []:
            for module in config.modules_to_not_convert:
                module_name = ".*" + module
                quant_config.set_local(module_name, TEQConfig(dtype="fp32"))
        logger.info(f"Do TEQ algorithm with config {quant_config}")
        run_fn = default_run_fn
        run_args = (
            config.tokenizer,
            config.dataset,
            config.seq_len,  # max_length
            config.n_samples,  # n_samples
            config.batch_size,  # batch_size
            config.quant_method.value,  # algo
        )
        example_inputs = torch.ones([1, 512], dtype=torch.long).to(device)
        model = prepare(model=model, quant_config=quant_config, example_inputs=example_inputs)
        run_fn(model, *run_args)
        model = convert(model)
    else:
        assert False, "The Supported algorithm are RTN, AWQ, TEQ, GPTQ"

    if device == "xpu" or device == torch.device("xpu"):
        logger.warning("The recommended ipex version is higher than 2.3.10 for xpu device.")

    model.eval()
    if not for_inference:
        # return inc model
        if config.use_layer_wise and not (model.device == device or model.device.type == device):
            logger.warning(
                "Do not convert device to avoid out of memory. Recommend using saved quantized model to inference."
            )
            return model
        return model.to(device)

    q_model = replace_linear(model, None, None, config, device=device)

    if config.use_layer_wise and not (q_model.device == device or q_model.device.type == device):
        logger.warning(
            "Do not convert device to avoid out of memory. Recommend using saved quantized model to inference."
        )
        return q_model

    return q_model.to(device)


def convert_to_GPTQ_checkpoints(model, quantization_config):
    from intel_extension_for_pytorch.nn.modules import WeightOnlyQuantizedLinear as ipex_cpu_linear

    from neural_compressor.torch.algorithms.weight_only.modules import INCWeightOnlyLinear
    from neural_compressor.torch.utils import set_module

    dtype = "int4" if quantization_config.bits == 4 else "int8"
    bits = quantization_config.bits
    group_size = quantization_config.group_size
    zp = False if quantization_config.sym else True
    scale_dtype = quantization_config.scale_dtype
    desc_act = (True if hasattr(quantization_config, "desc_act") else False,)

    for name, module in model.named_modules():
        if isinstance(module, ipex_cpu_linear):
            in_features = module.in_features
            out_features = module.out_features
            new_module = INCWeightOnlyLinear(
                in_features,
                out_features,
                dtype=dtype,
                bits=bits,
                group_size=group_size,
                zp=zp,
                bias=True if hasattr(module, "bias") else False,
                scale_dtype=scale_dtype,
                g_idx=desc_act,
                use_optimum_format=True,
            )

            new_module.bits = 8
            new_module.n_pack = 32 // 8
            qweight = (
                new_module.pack_tensor_with_numpy(module._op_context.to_public(module._op_context.get_weight()))
                .t()
                .contiguous()
            )
            new_module.bits = bits
            new_module.n_pack = 32 // bits
            scales = module._op_context.get_scales().t().contiguous()
            bias = module._op_context.get_bias()
            qzeros = module._op_context.get_zero_points().t().to(torch.uint8)
            # For group_size = -1, the dimensions of scale and qzeros will be 1
            if len(scales.shape) == 1:
                scales = scales.unsqueeze(0)
            if len(qzeros.shape) == 1:
                qzeros = qzeros.unsqueeze(0)
            qzeros = new_module.pack_tensor_with_numpy(qzeros - 1).contiguous()
            g_idx = module._op_context.get_g_idx()

            new_module.qweight = qweight
            new_module.scales = scales
            new_module.qzeros = qzeros
            if g_idx is not None:
                new_module.g_idx = g_idx.contiguous()
            if bias is not None:
                new_module.bias = bias.contiguous()

            set_module(model, name, new_module)
    return model


def make_contiguous(model):
    for param in model.parameters():
        if param.data.ndimension() > 1:
            param.data = param.data.contiguous()


def save_low_bit_for_inc(self, save_directory: Union[str, os.PathLike], push_to_hub: bool = False, **kwargs):

    assert hasattr(self, "quantization_config"), "Detected this model is not a low-bit model."

    if os.path.isfile(save_directory):
        logger.error(f"Provided path ({save_directory}) should be a directory, not a file")
        return

    os.makedirs(save_directory, exist_ok=True)
    # use transformers original `save_pretrained` function
    del self.save_pretrained
    make_contiguous(self)

    self.save_pretrained(save_directory=save_directory, push_to_hub=push_to_hub, **kwargs)
    self.save_pretrained = types.MethodType(save_low_bit_for_inc, self)
    # We conveniently save all the keys of the model to have them on hand,
    # so that when using 'low_cpumem load',
    # it's not necessary to load the entire model to extract its keys
    # and we can avoid gc not triggered potentially.
    all_checkpoint_keys = {"all_checkpoint_keys": list(self.state_dict().keys())}
    json_file_path = os.path.join(save_directory, "all_checkpoint_keys.json")
    with open(json_file_path, "w") as json_file:
        json.dump(all_checkpoint_keys, json_file)
    if push_to_hub:
        use_auth_token = kwargs.pop("use_auth_token", None)

        if use_auth_token is not None:
            logger.warning.warn(
                "The `use_auth_token` argument is deprecated and will be removed in v5 of Transformers.",
                FutureWarning,
            )

            token = use_auth_token
            if token is not None:
                raise ValueError(
                    "`token` and `use_auth_token` are both specified. Please set only the argument `token`."
                )

            if token is not None:
                kwargs["token"] = token
        commit_message = kwargs.pop("commit_message", None)
        repo_id = kwargs.pop("repo_id", save_directory.split(os.path.sep)[-1])
        repo_id = self._create_repo(repo_id, **kwargs)
        files_timestamps = self._get_files_timestamps(save_directory)
        self._upload_modified_files(
            save_directory,
            repo_id,
            files_timestamps,
            commit_message=commit_message,
            token=kwargs.get("token"),
        )
    self.quantization_config.save_pretrained(save_directory, **kwargs)


def save_low_bit(self, save_directory: Union[str, os.PathLike], push_to_hub: bool = False, **kwargs):

    assert hasattr(self, "quantization_config"), "Detected this model is not a low-bit model."

    if os.path.isfile(save_directory):
        logger.error(f"Provided path ({save_directory}) should be a directory, not a file")
        return

    os.makedirs(save_directory, exist_ok=True)
    # use transformers original `save_pretrained` function
    del self.save_pretrained
    make_contiguous(self)

    if self.device == "cpu" or self.device == torch.device("cpu"):
        convert_to_GPTQ_checkpoints(self, self.quantization_config)
    if self.device == "xpu" or (isinstance(self.device, torch.device) and self.device.type == "xpu"):
        from intel_extension_for_pytorch.nn.utils._quantize_convert import WeightOnlyQuantizedLinear

        for name, module in self.named_modules():
            if isinstance(module, WeightOnlyQuantizedLinear):
                if module.weight_transposed:
                    module.qweight.data = module.qweight.t_().contiguous()
                    module.scales.data = module.scales.t_().contiguous()
                    module.weight_transposed = False

    self.save_pretrained(save_directory=save_directory, push_to_hub=push_to_hub, **kwargs)
    self.save_pretrained = types.MethodType(save_low_bit, self)
    # We conveniently save all the keys of the model to have them on hand,
    # so that when using 'low_cpumem load',
    # it's not necessary to load the entire model to extract its keys
    # and we can avoid gc not triggered potentially.
    all_checkpoint_keys = {"all_checkpoint_keys": list(self.state_dict().keys())}
    json_file_path = os.path.join(save_directory, "all_checkpoint_keys.json")
    with open(json_file_path, "w") as json_file:
        json.dump(all_checkpoint_keys, json_file)
    if push_to_hub:
        use_auth_token = kwargs.pop("use_auth_token", None)

        if use_auth_token is not None:
            logger.warning.warn(
                "The `use_auth_token` argument is deprecated and will be removed in v5 of Transformers.",
                FutureWarning,
            )

            token = use_auth_token
            if token is not None:
                raise ValueError(
                    "`token` and `use_auth_token` are both specified. Please set only the argument `token`."
                )

            if token is not None:
                kwargs["token"] = token
        commit_message = kwargs.pop("commit_message", None)
        repo_id = kwargs.pop("repo_id", save_directory.split(os.path.sep)[-1])
        repo_id = self._create_repo(repo_id, **kwargs)
        files_timestamps = self._get_files_timestamps(save_directory)
        self._upload_modified_files(
            save_directory,
            repo_id,
            files_timestamps,
            commit_message=commit_message,
            token=kwargs.get("token"),
        )
    self.quantization_config.save_pretrained(save_directory, **kwargs)


def repack_awq_and_load_state_dict(
    model, resolved_archive_file, loaded_state_dict_keys, quantization_config, is_sharded
):
    from transformers.modeling_utils import load_state_dict

    bits = quantization_config.bits
    group_size = quantization_config.group_size

    state_dict = {}
    if isinstance(resolved_archive_file, str):
        resolved_archive_file = [resolved_archive_file]
    assert isinstance(resolved_archive_file, list), "Please check if the loading weight is shared."
    for shard_file in resolved_archive_file:
        assert shard_file.endswith("safetensors"), "Please check the loading weight saved format."
        state_dict.update(load_state_dict(shard_file))
        assert len(state_dict.keys()) > 0, "Please check the state_dict loading."
    for name, module in model.named_modules():
        if isinstance(module, INCWeightOnlyLinear):
            assert name + ".qweight" in loaded_state_dict_keys, f"Please check the state_dict key { name + '.qweight'}"
            assert name + ".qzeros" in loaded_state_dict_keys, f"Please check the state_dict key {name + '.qzeros'}"
            assert name + ".scales" in loaded_state_dict_keys, f"Please check the state_dict key { name + '.scales'}"
            if name + ".scales" in loaded_state_dict_keys:
                awq_qweight = state_dict[name + ".qweight"]
                awq_qzeros = state_dict[name + ".qzeros"]
                awq_scales = state_dict[name + ".scales"]
                qweight, qzeros, awq_scales = repack_awq_to_optimum_format(
                    awq_qweight, awq_qzeros, awq_scales, bits, group_size
                )
                state_dict[name + ".qweight"] = qweight
                state_dict[name + ".qzeros"] = qzeros
                state_dict[name + ".scales"] = awq_scales

    model.load_state_dict(state_dict, strict=False, assign=True)

    return model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class and helper functions for registering scaling methods."""

from typing import Dict, Optional, Any
from abc import abstractmethod, ABC
import torch
from neural_compressor.torch.algorithms.fp8_quant.model_configs import (
    ModuleConfig,
    SCALING_METHODS_TABLE,
    import_external_scaling_methods,
)

__all__ = [
    "ScalingMethodBase",
    "register_scaling_methods",
]


class ScalingMethodBase(ABC):
    """Base class for scaling methods.

    This class defines the interface for generating scaling methods for patched modules.
    """

    @staticmethod
    @abstractmethod
    def generate_op_scale_method(
        mod: torch.nn.Module,
        measurement: Optional[ModuleConfig] = None,
        params: Optional[Dict[str, Any]] = None,
        device: Optional[str] = "hpu",
        **kwargs: Dict[str, Any],
    ) -> ModuleConfig:
        raise NotImplementedError("`generate_op_scale_method` is not implemented")

    @staticmethod
    @abstractmethod
    def op_scales_to_mod_config(
        mod: torch.nn.Module,
        scales: ModuleConfig,
        params: Dict[str, Any],
        device: Optional[str] = "hpu",
        **kwargs: Dict[str, Any],
    ) -> ModuleConfig:
        raise NotImplementedError("`op_scales_to_mod_config` is not implemented")


def register_scaling_methods(patch_module_type: str, scaling_method_name: str):
    """Register a scaling method for a patched module.

    Args:
        patch_module_type (str): The type of the patched module.
        scaling_method_name (str): The name of the scaling method.

    Examples:
        @register_scaling_methods("linear", "unit_scale")
        class ScaleByMax(ScalingMethodBase):
            ...
    """

    def decorate(cls: ScalingMethodBase):
        SCALING_METHODS_TABLE[scaling_method_name] = (patch_module_type, cls)
        import_external_scaling_methods()
        return cls

    return decorate



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from enum import Enum, auto


class OP_TYPE(Enum):
    # class per custom fp8 ops used in patched modules logic
    LINEAR_GEMM = auto()
    MATMUL_GEMM = auto()
    SOFTMAX = auto()
    BLOCK_SOFTMAX_CONST_MAX = auto()
    CONV = auto()
    FSDPA = auto()
    CAST_TO_FP8 = auto()
    CAST_FROM_FP8 = auto()
    DYNAMIC_MOE = auto()
    DYNAMIC_MOE_FUSED_WEIGHTS = auto()
    QUANT = auto()
    DEQUANT = auto()
    QUANT_PC = auto()
    DEQUANT_PC = auto()


class QuantizedFuncWrapperBase(ABC):
    """
    Base class for wrapping calls to custom fp8 ops.
    The concrete class object is created in patched module init in call to get_<device>_quantized_func_wrapper.
    Concrete class should define get_default_quantized_func method.
    Concrete class may override base class methods in case custom op logic is unique, see examples in concrete
    classes implementation.
    """
    @abstractmethod
    def get_default_quantized_func(self):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        return self._quantized_func_(*args, **kwargs)


class QuantizedFuncWrapperFactory():
    """
    A Factory object to create func wrappers objects.
    This is a singleton and it creates single object per quantized func wrapper class.
    This is done to avoid unnecessary duplication of quantized func wrapper objects since, since they are all identical.
    """

    __is_initialized = False

    # using a global map and the below get_quantized_func_wrapper method,
    # ensures only a single object of each quantized func wrapper concrete class will be created.
    __quantized_func_wrapper_instances = {}

    __device_func_wrappers_mapping = {}

    def __new__(cls):
        raise TypeError("The {} class can't be instantiated directly".format(cls.__name__))

    @classmethod
    def initialize(cls, device_quantized_func_wrapper_dict):
        if not cls.__is_initialized:
            if not device_quantized_func_wrapper_dict:
                raise ValueError("Expected non empty device_quantized_func_wrapper_dict")
            cls.__device_func_wrappers_mapping = dict(device_quantized_func_wrapper_dict)
            cls.__is_initialized = True


    @classmethod
    def get_quantized_func_wrapper_object(cls, op_type, scale_format, is_dynamic=False):
        if op_type not in cls.__quantized_func_wrapper_instances:
            quantized_wrapper_class = cls.__device_func_wrappers_mapping[op_type]
            cls.__quantized_func_wrapper_instances[op_type] = quantized_wrapper_class(scale_format, is_dynamic)

        return cls.__quantized_func_wrapper_instances[op_type]

    @classmethod
    def clear(cls):
        cls.__quantized_func_wrapper_instances.clear()
        cls.__is_initialized = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MX quantization."""

from collections import OrderedDict

import torch
from torch.nn import functional as F

from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.utils import logger, set_module

from .utils import quantize_elemwise_op, quantize_mx_op


class MXLinear(torch.nn.Linear):
    """Linear for MX data type."""

    def __init__(
        self,
        in_features,
        out_features,
        bias=False,
        mx_specs=None,
        name=None,
    ):
        """Initialization function."""
        self.mx_none = mx_specs is None

        self.name = name
        self.mx_specs = mx_specs
        super().__init__(in_features, out_features, bias)

    def apply_mx_specs(self):
        """Apply MX data type to weight."""
        if self.mx_specs is not None:
            if self.mx_specs.out_dtype != "float32":
                self.weight.data = quantize_elemwise_op(self.weight.data, mx_specs=self.mx_specs)

                if self.bias is not None:
                    self.bias.data = quantize_elemwise_op(self.bias.data, mx_specs=self.mx_specs)

            # MX quantize everything along input size
            self.weight.data = quantize_mx_op(
                self.weight.data,
                self.mx_specs.w_dtype,
                self.mx_specs.round_method,
                self.mx_specs.blocksize,
                axes=[-1],
            )

    def forward(self, input):
        """Forward function."""
        if self.mx_none:
            return super().forward(input)

        # element-wise quantize for input
        if self.mx_specs.out_dtype != "float32":
            input = quantize_elemwise_op(input, mx_specs=self.mx_specs)

        if not self.mx_specs.weight_only:
            # MX quantize everything along input size
            input = quantize_mx_op(
                input,
                self.mx_specs.act_dtype,
                self.mx_specs.round_method,
                self.mx_specs.blocksize,
                axes=[-1],
            )
        # compute output
        output = F.linear(input, self.weight)
        if self.mx_specs.out_dtype != "float32":
            output = quantize_elemwise_op(output, mx_specs=self.mx_specs)

        if self.bias is not None:
            output = output + self.bias
            if self.mx_specs.out_dtype != "float32":
                output = quantize_elemwise_op(output, mx_specs=self.mx_specs)

        return output


class MXQuantizer(Quantizer):
    """Quantizer of MX data type."""

    def __init__(self, quant_config: OrderedDict = {}):
        """Init a MXQuantizer object.

        Args:
            quant_config (OrderedDict, optional): quantization config for ops. Defaults to {}.
        """
        super().__init__(quant_config)

    @torch.no_grad()
    def prepare(self, model, *args, **kwargs):
        """Prepares a given model for quantization.

        Will return model directly in MX quant algorithm.

        Args:
            model (torch.nn.Module): The model to be prepared.
        """
        return model

    @torch.no_grad()
    def convert(
        self,
        model,
        **kwargs,
    ):
        """Quant the model with mx data type.

        Args:
            model: torch module

        Returns:
            model: fake quantized torch module
        """
        assert isinstance(model, torch.nn.Module), "only support torch module"
        supported_layers = ["Linear"]
        config = self.quant_config
        for name, m in model.named_modules():
            if type(m).__name__ not in supported_layers:
                continue
            if (name, type(m).__name__) not in config:  # pragma: no cover
                continue
            logger.debug(f"MX quantized module:{name, m}")
            log_msg = (
                f"MX quantization config: w_dtype={config[(name, type(m).__name__)].w_dtype}, "
                + "config[(name, type(m).__name__)].act_dtype, "
                + f"out_dtype={config[(name, type(m).__name__)].out_dtype}"
            )
            logger.debug(log_msg)
            tmp_stat = m.state_dict()
            new_module = MXLinear(
                m.in_features,
                m.out_features,
                bias=m.bias is not None,
                mx_specs=config[(name, type(m).__name__)],
                name=name,
            )
            new_module.load_state_dict(tmp_stat)
            new_module.apply_mx_specs()
            if name == "":
                return new_module
            else:
                set_module(model, name, new_module)
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
#
# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TensorQuantizer Module."""

import torch
from torch import nn

try:
    from auto_round.data_type import get_quant_func
except ImportError:
    get_quant_func = None


class TensorQuantizer(nn.Module):
    """Tensor quantizer module."""

    def __init__(
        self,
        data_type="mx_fp8",
        bits=8,
        block_size=32,
        sym=True,
        if_quant=True,
        learn_exponent=False,
        amax=None,
        scale_shape=None,
        device=None,
    ):
        """Initialize quantizer and set up required variables."""
        super().__init__()
        self.amax = amax
        self.data_type = data_type
        self.num_bits = bits
        self.block_size = block_size
        self.sym = sym
        self._if_quant = if_quant
        self.learn_exponent = learn_exponent
        self._dequantize = False
        self._input_dtype = None
        self._fake_quant = True

        # enable quantizer
        self.enable()

        assert (
            get_quant_func is not None
        ), "The quantization function is imported from AutoRound, please install it. 'pip install auto-round'"

        # self.data_type will be overided 'mx_fp' -> 'mx_fp8'
        self.quant_func, self.data_type = get_quant_func(self.data_type, self.num_bits, self.sym)

        if scale_shape is not None:
            # E8M0 scales (exponent)
            self.register_buffer(
                "scale",
                torch.empty(scale_shape[0], scale_shape[1] // self.block_size, dtype=torch.uint8, device=device),
            )
            self.save_scale = True
        else:
            self.save_scale = False

    def forward(self, inputs: torch.Tensor):
        """Apply tensor_quant function to inputs.

        Args:
            inputs: A Tensor of type float32/float16/bfloat16.

        Returns:
            outputs: A Tensor of type output_dtype
        """

        if self._disabled or (not self._if_quant):
            self._input_dtype = inputs.dtype
            return inputs

        x = inputs
        if not x.is_contiguous():
            x = x.contiguous()

        if self.fake_quant:
            q = self._fake_quantize(x)[0]
        else:
            # TODO: add implementation
            q = self._real_quantize(x)

        return q.to(inputs.dtype)

    def _fake_quantize(self, inputs: torch.Tensor):
        """Fake quantization."""

        # the shared_exp can be trainable
        if self.learn_exponent:
            q, shared_exp, _ = self.quant_func(
                inputs,
                bits=self.num_bits,
                group_size=self.block_size,
                data_type=self.data_type,
            )
        else:
            # wrapper no_grad, because the function includes extra trainable variables
            with torch.no_grad():
                q, shared_exp, _ = self.quant_func(
                    inputs,
                    bits=self.num_bits,
                    group_size=self.block_size,
                    data_type=self.data_type,
                )

            # simple STE, since we add no_grad in the quant function
            q = q.detach() + (inputs - inputs.detach())

        if self.save_scale:
            # TODO: PACK uint8
            self.scale.data.copy_(shared_exp.detach())

        return q, shared_exp

    def _real_quantize(self, inputs: torch.Tensor):
        raise NotImplementedError("This method hasn't be implemented.")

    @property
    def fake_quant(self):
        """Return True if fake quantization is used."""
        return self._fake_quant

    def disable(self):
        """Bypass the module."""
        self._disabled = True

    def enable(self):
        """Enable the module."""
        self._disabled = False

    def weight_pack(self, weight, scale):
        """Pack weight and scale when saving."""
        original_shape = weight.shape

        # TODO: support more quantization format
        if self.data_type == "mx_fp8":
            qweight = (weight.reshape(-1, self.block_size) / torch.exp2(scale.float()).reshape(-1, 1)).to(
                torch.float8_e4m3fn
            )

            e8m0_scale = (scale + 127).to(torch.uint8)
            return qweight.reshape(original_shape), e8m0_scale.reshape(original_shape[0], -1)

        if self.data_type == "mx_fp4":
            qweight = weight.reshape(-1, self.block_size) / torch.exp2(scale.float())

            from auto_round.export.export_to_autoround.qlinear_fp import pack_fp4_to_uint8

            qweight_packed = pack_fp4_to_uint8(qweight)

            e8m0_scale = (scale + 127).to(torch.uint8)
            return qweight_packed.reshape(original_shape[0], original_shape[1] // 2), e8m0_scale.reshape(
                original_shape[0], -1
            )

    def __repr__(self):
        if self._disabled or not self._if_quant:
            return "TensorQuantizer(disabled)"

        qformat_str = f"({self.data_type}) format"
        bits_str = f"({self.num_bits}) bit"

        if self.block_size:
            bs_str = f"block_size={self.block_size}"
        else:
            bs_str = "block_size=None"

        # amax
        amax_str = f"amax={self.amax}" if self.amax is not None else "amax=?"
        # fake / real
        mode_str = "fake" if self._fake_quant else "real"
        # sym
        sym_str = "sym" if self.sym else "asym"
        # quant enable
        qflag = "quant" if self._if_quant else "no-quant"

        return f"TensorQuantizer({qformat_str} {bits_str} {mode_str} {bs_str}, {amax_str} {qflag})"



# Copyright (c) 2023-2024 Mobiusml and Intel Corporation

# This code is based on Mobiusml's HQQ library. It has been modified
# from its original forms to simplify and adapt it for use in
# the Intel® Neural Compressor.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE: the original `Quantizer` has been modified to `HQQTensorHandle`
# and `QTensor` to decouple the data structure and the quantization logic.
"""The HQQ modules."""


from typing import Any, Dict, Mapping, Tuple

import torch

from neural_compressor.common.utils import dump_elapsed_time
from neural_compressor.torch.utils import logger
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

from .bitpack import Packer
from .config import HQQModuleConfig, QTensorConfig, default_hqq_module_config, hqq_global_option
from .optimizer import optimize_weights_proximal
from .qtensor import QTensor, QTensorMetaInfo

__all__ = [
    "HQQTensorHandle",
    "HQQLinear",
]


class HQQTensorHandle:
    """HQQ Tensor Handle to quantize and dequantize the tensor."""

    # Refactored the code from https://github.com/mobiusml/hqq.

    # Store meta-data (we invert the scale for dequantization)
    SUPPORTED_BITS = [8, 4, 3, 2]
    optimize_weights = optimize_weights_proximal

    @classmethod
    def quantize(cls, float_tensor, tensor_quant_config: QTensorConfig = None):
        """Quantizes a given float tensor using the specified tensor quantization configuration.

        Args:
            float_tensor (torch.Tensor): The float tensor to be quantized.
            tensor_quant_config (QTensorConfig, optional): The tensor quantization configuration. Defaults to None.

        Returns:
            torch.Tensor: The quantized tensor.
        """
        q_weight, q_tensor_meta = cls._quantize(
            tensor=float_tensor,
            tensor_quant_config=tensor_quant_config,
        )
        q_weight = cls._create_q_tensor(q_weight, q_tensor_meta)
        return q_weight

    @classmethod
    def dequantize(cls, q_weight: "QTensor") -> torch.Tensor:
        """Dequantizes the QTensor into a float tensor.

        Args:
            q_weight (QTensor): The quantized weight tensor.

        Returns:
            torch.Tensor: The dequantized float tensor.
        """
        meta = q_weight.meta_info.to_dict()
        meta["zero"] = q_weight.zero
        meta["scale"] = q_weight.scale
        return cls._dequantize(q_weight.val, meta)

    @classmethod
    def _create_q_tensor(cls, weight, meta) -> "QTensor":
        scale = meta["scale"]
        zero = meta["zero"]
        meta_info = QTensorMetaInfo(
            nbits=meta["nbits"],
            group_size=meta["group_size"],
            shape=meta["shape"],
            axis=meta["axis"],
            packing=meta["packing"],
        )
        return QTensor(weight, scale, zero, meta_info)

    @classmethod
    def _quantize(cls, tensor, tensor_quant_config: QTensorConfig = None):
        nbits = tensor_quant_config.nbits
        channel_wise = tensor_quant_config.channel_wise
        group_size = tensor_quant_config.group_size if tensor_quant_config.group_size != -1 else None
        optimize = tensor_quant_config.optimize
        round_zero = tensor_quant_config.round_zero
        axis = 0  # *Note did not exposed to the user
        bitpack = tensor_quant_config.pack

        assert nbits in cls.SUPPORTED_BITS, "nbits=" + str(nbits) + " not supported."
        assert axis in [0, 1], "axis should be either 0 or 1, but got {}".format(axis)
        if group_size is not None:
            assert tensor.numel() % group_size == 0, (
                "group_size should be divisible by the total tensor dimensions. shape: "
                + str(tensor.shape)
                + ", group_size: "
                + str(group_size)
            )

        W = tensor.float()
        shape = W.shape

        # Reshape for grouping
        if (group_size is not None) and channel_wise:
            W = W.reshape([-1, group_size]) if (axis == 1) else W.reshape([group_size, -1])

        # Get min/max values
        if not channel_wise:
            _min, _max = W.min(), W.max()
            optimize = False
        else:
            _min = W.min(axis=axis, keepdim=True)[0]
            _max = W.max(axis=axis, keepdim=True)[0]

        max_v = 2**nbits - 1
        min_v = 0
        min_max = [min_v, max_v]

        # Note: here we work with the inverse of the scale to avoid division and quantize instead via W*scale + zero,
        # the scale is inverted later on.
        scale = (max_v / (_max - _min)).clamp(max=2e4)  # clamp to avoid half-precision problems
        zero = -_min * scale

        # Round zero as in: https://github.com/casper-hansen/AutoAWQ/blob/main/awq/quantize/quantizer.py#L42C9-L42C14
        if round_zero:
            zero = torch.round(zero)

        # Fine-tune weights
        if optimize:
            scale, zero = cls.optimize_weights(tensor=W, scale=scale, zero=zero, min_max=min_max, axis=axis)

        # Quantize
        scale, zero = (
            scale.clone(),
            zero.clone(),
        )  # Necessary for fake quantization backprop
        W_q = torch.round(W * scale + zero).clamp(min_max[0], min_max[1])

        # Store meta-data (we invert the scale for dequantization)
        meta = {
            "nbits": nbits,
            "group_size": group_size,
            "shape": shape,
            "scale": 1.0 / scale,
            "zero": zero,
            "axis": axis,
            "packing": bitpack,
        }

        # Pack bits
        if bitpack:
            W_q = Packer.get_pack_fn(meta["nbits"])(W_q)
        else:
            W_q = W_q.to(tensor.dtype)
            meta["packing"] = None

        # cleanup
        del W, _min, _max
        auto_detect_accelerator().empty_cache()

        return W_q, meta

    @classmethod
    def _dequantize(cls, W_q, meta):
        # Main dequantization: bit_unpacking > (W_q - z)*s > reshape
        if meta["packing"]:
            W_r = Packer.get_unpack_fn(meta["nbits"])(W_q)
            if hqq_global_option.use_half:
                W_r = W_r.half()
            if (meta["group_size"] is not None) and (meta["nbits"] == 3):
                W_r = W_r[: meta["group_size"]] if (meta["axis"] == 0) else W_r[:, : meta["group_size"]]
        else:
            if hqq_global_option.use_half:
                W_r = W_q.half()
        # TODO: double check the correctness, the official impl is also error...
        W_r = ((W_r - meta["zero"]) * meta["scale"]).reshape(meta["shape"])
        return W_r


class HQQLinear(torch.nn.Linear):
    """HQQ Linear module."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        q_weight: QTensor = None,
        device=None,
        dtype=None,
    ) -> None:
        """Init a HQQ linear."""
        super().__init__(in_features, out_features, bias, device, dtype)
        self.q_weight = q_weight
        self.quantized = q_weight is not None

    @dump_elapsed_time("Quantize linear module into HQQ module.")
    def quantize_weight(
        self,
        W: torch.Tensor,
        quant_config: HQQModuleConfig = default_hqq_module_config,
    ) -> Tuple[torch.Tensor, Dict[str, Any]]:
        """Quantizes the weight using HQQ.

        Args:
            W (torch.Tensor): The weight tensor to be quantized.
            quant_config (HQQModuleConfig, optional): The quantization configuration.
                Defaults to default_hqq_module_config.

        Returns:
            Tuple[torch.Tensor, Dict[str, Any]]: A tuple containing the quantized weight tensor
                and a dictionary of additional information.
        """
        weight_quant_config, scale_quant_config, zero_quant_config = (
            quant_config.weight,
            quant_config.scale,
            quant_config.zero,
        )
        need_quant_scale = scale_quant_config is not None
        need_quant_zero = zero_quant_config is not None

        self.in_features, self.out_features = W.t().shape

        # Quantize weight
        q_weight = HQQTensorHandle.quantize(float_tensor=W, tensor_quant_config=weight_quant_config)
        self.q_weight = q_weight

        # * The dequantization process only happens in the first forward pass.
        # * It will change the `q_weight` but faster.
        # * we should not save the state after doing the forward.
        if need_quant_scale:  # Quantize scale
            q_scale_tensor = HQQTensorHandle.quantize(
                float_tensor=self.q_weight.scale, tensor_quant_config=scale_quant_config
            )
            self.q_weight.scale = q_scale_tensor
        if need_quant_zero:  # Quantize zero
            q_zero_tensor = HQQTensorHandle.quantize(
                float_tensor=self.q_weight.zero,
                tensor_quant_config=zero_quant_config,
            )
            self.q_weight.zero = q_zero_tensor
        self.quantized = True

    def dequantize_weight(self):
        """Dequantize the weight tensor."""
        assert self.quantized, "model was not quantized"
        # TODO: move below logic into `HQQTensorHandle`
        if self.q_weight.is_scale_quantized():
            scale_qdq = HQQTensorHandle.dequantize(self.q_weight.scale)
            self.q_weight.scale = scale_qdq

        if self.q_weight.is_zero_quantized():
            zero_qdq = HQQTensorHandle.dequantize(self.q_weight.zero)
            self.q_weight.zero = zero_qdq

        W_qdq = HQQTensorHandle.dequantize(self.q_weight)
        return W_qdq

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass of the HQQ linear module."""
        out = torch.matmul(input, self.dequantize_weight().t())
        if self.bias is not None:
            out += self.bias
        return out

    @classmethod
    def from_float(
        cls,
        float_module: torch.nn.Linear,
        quant_config: HQQModuleConfig = default_hqq_module_config,
    ):
        """Create a new HQQModule instance from a floating-point linear.

        Args:
            float_module (torch.nn.Linear): The floating-point module to convert.
            quant_config (HQQModuleConfig, optional): The quantization configuration.
                Defaults to default_hqq_module_config.

        Returns:
            HQQModule: The converted HQQModule instance.
        """
        # Create the new module with a toy size to ensure initialization is fast
        fake_in_features, fake_out_features = 8, 8
        new_mod = cls(
            fake_in_features,
            fake_out_features,
            bias=float_module.bias is not None,
        )
        new_mod.requires_grad_ = False
        # Construct the q weight from float weight
        new_mod.quantize_weight(float_module.weight, quant_config=quant_config)
        # Update the linear module attributes
        new_mod.in_features = float_module.in_features
        new_mod.out_features = float_module.out_features
        new_mod.weight = None
        new_mod.bias = float_module.bias
        if hqq_global_option.use_half and new_mod.bias is not None:
            new_mod.bias = torch.nn.Parameter(float_module.bias.half())
        # TODO: refine it to support cuda/hpu/cpu
        device_to_use = next(float_module.parameters()).device
        if hqq_global_option.use_half:
            new_mod.q_weight = new_mod.q_weight.half()
        new_mod.to(device_to_use)
        new_mod.q_weight.to(device_to_use)
        # !!! Delete the float explicitly to save memory
        del float_module
        return new_mod

    def state_dict(self, *args, **kwargs):  # nn.Module override compatible
        """Returns a dictionary containing the state of the module.

        The state dictionary contains the weights of the `q_weight` attribute.
        If the `bias` attribute is not None, it is also included in the state dictionary.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: A dictionary containing the state of the module.
        """
        state_dict = self.q_weight.to_state_dict()
        if self.bias is not None:
            state_dict["bias"] = self.bias
        if "destination" in kwargs and "prefix" in kwargs:
            for key, value in state_dict.items():
                kwargs["destination"][kwargs["prefix"] + key] = value
        return state_dict

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        all_expected_keys = ["val", "scale_quantized", "zero_quantized", "meta_info"]
        if self.bias is not None:
            all_expected_keys.append("bias")

        for key in all_expected_keys:
            if prefix + key not in state_dict:
                missing_keys.append(key)
        if missing_keys:
            return  # Can't load weights if either weight or meta is missing

        cur_state_dict = {}
        for key in all_expected_keys:
            cur_state_dict[key] = state_dict.pop(prefix + key)

        unexpected_keys += state_dict.keys()
        self._assign_state_dict(cur_state_dict, strict)

    def _assign_state_dict(self, state_dict: Mapping[str, Any], strict: bool = True, assign: bool = False):
        _scale_quantized = state_dict["scale_quantized"]
        _zero_quantized = state_dict["zero_quantized"]
        scale_state = state_dict["meta_info"]["scale"]
        zero_state = state_dict["meta_info"]["zero"]
        if _scale_quantized:
            scale = HQQTensorHandle._create_q_tensor(scale_state["val"], scale_state["meta_info"])
        else:
            scale = state_dict["meta_info"]["scale"]
        if _zero_quantized:
            zero = HQQTensorHandle._create_q_tensor(zero_state["val"], zero_state["meta_info"])
        else:
            zero = state_dict["meta_info"]["zero"]
        meta = state_dict["meta_info"]
        meta["scale"] = scale
        meta["zero"] = zero
        self.q_weight = HQQTensorHandle._create_q_tensor(state_dict["val"], meta)
        if self.bias is not None:
            self.bias = state_dict["bias"]
        self.quantized = True
        return self
